#include "stdafx.h"
#include "RPServer.h"
#include "RPServerDlg.h"
#include "Comm.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

//***************************************************************************
// CCommPara class
CCommPara::CCommPara(void)
{
	m_Name = "COM1";
	m_Baud = 4800;
	m_Length = 1;
	m_Stop = 1;
	m_Parity = NOPARITY;
	m_Flow = 0;
	m_PTT = 0;
}
//---------------------------------------------------------------------------
void __fastcall CCommPara::LoadProfile(CWinApp *pApp)
{
	m_Name = pApp->GetProfileString("Port", "Name", m_Name);
	m_Baud = pApp->GetProfileInt("Port", "Baud", m_Baud);
	m_Length = pApp->GetProfileInt("Port", "Length", m_Length);
	m_Stop = pApp->GetProfileInt("Port", "Stop", m_Stop);
	m_Parity = pApp->GetProfileInt("Port", "Parity", m_Parity);
	m_Flow = pApp->GetProfileInt("Port", "Flow", m_Flow);
	m_PTT = pApp->GetProfileInt("Port", "PTT", m_PTT);
}
//---------------------------------------------------------------------------
void __fastcall CCommPara::SaveProfile(CWinApp *pApp)
{
	pApp->WriteProfileString("Port", "Name", m_Name);
	pApp->WriteProfileInt("Port", "Baud", m_Baud);
	pApp->WriteProfileInt("Port", "Length", m_Length);
	pApp->WriteProfileInt("Port", "Stop", m_Stop);
	pApp->WriteProfileInt("Port", "Parity", m_Parity);
	pApp->WriteProfileInt("Port", "Flow", m_Flow);
	pApp->WriteProfileInt("Port", "PTT", m_PTT);
}
//***************************************************************************
// CFifo class
//---------------------------------------------------------------------------
CFifo::CFifo(void)
{
	m_cc = m_rc = m_wc = 0;
}
//---------------------------------------------------------------------------
// Xbhs͌Ă΂Ȃ
void __fastcall CFifo::Clear(void)
{
	m_cc = m_rc = m_wc = 0;
}
//---------------------------------------------------------------------------
void __fastcall CFifo::PutData(BYTE c)
{
	if( IsBufferFull() ) return;
	m_bf[m_wc] = c;
	m_wc++;
	if( m_wc >= BFIFOMAX ) m_wc = 0;
	::InterlockedIncrement(&m_cc);
}
//---------------------------------------------------------------------------
BYTE __fastcall CFifo::GetData(void)
{
	if( !m_cc ) return 0;
	BYTE c = m_bf[m_rc];
	m_rc++;
	if( m_rc >= BFIFOMAX ) m_rc = 0;
	::InterlockedDecrement(&m_cc);
	return c;
}
//---------------------------------------------------------------------------
int __fastcall CFifo::GetCount(void)
{
	return m_cc;
}

//***************************************************************************
// CComm class
//---------------------------------------------------------------------------
CComm::CComm()
{
	m_hWnd = NULL;
	m_pThread = NULL;
	m_hPort = INVALID_HANDLE_VALUE;
	m_Terminated = FALSE;
#if !USE_POLLING
	memset(&m_OverlapTX, 0, sizeof(m_OverlapTX));
	memset(&m_OverlapRX, 0, sizeof(m_OverlapRX));
#endif
}
//---------------------------------------------------------------------------
CComm::~CComm()
{
	Close();
}
//---------------------------------------------------------------------------
void __fastcall CComm::ClearBuffer(void)
{
	m_FifoRX.Clear();
	m_FifoTX.Clear();
}
//---------------------------------------------------------------------------
BOOL __fastcall CComm::Open(HWND hWnd, const CCommPara *pPara)
{
	Close();
	m_Terminated = FALSE;
	m_pPara = pPara;
	m_hWnd = hWnd;
	ClearBuffer();
	m_hPort = ::CreateFile( pPara->m_Name,
					GENERIC_READ | GENERIC_WRITE,
					0, NULL,
					OPEN_EXISTING,
#if USE_POLLING
					FILE_ATTRIBUTE_NORMAL,
#else
					FILE_FLAG_OVERLAPPED,
#endif
					NULL
	);
	if( m_hPort == INVALID_HANDLE_VALUE ) return FALSE;
	if( ::SetupComm( m_hPort, DWORD(256), DWORD(256) ) == FALSE ){
		::CloseHandle(m_hPort);
		m_hPort = INVALID_HANDLE_VALUE;
		return FALSE;
	}
	::PurgeComm( m_hPort, PURGE_TXABORT | PURGE_RXABORT | PURGE_TXCLEAR | PURGE_RXCLEAR );

	COMMTIMEOUTS TimeOut;

#if USE_POLLING
	TimeOut.ReadIntervalTimeout = 0xffffffff;
	TimeOut.ReadTotalTimeoutMultiplier = 0;
	TimeOut.ReadTotalTimeoutConstant = 0;
	TimeOut.WriteTotalTimeoutMultiplier = 0;
	TimeOut.WriteTotalTimeoutConstant = 20000;
#else
	TimeOut.ReadIntervalTimeout = 1000;
	TimeOut.ReadTotalTimeoutMultiplier = 1;
	TimeOut.ReadTotalTimeoutConstant = 20000;
	TimeOut.WriteTotalTimeoutMultiplier = 1;
	TimeOut.WriteTotalTimeoutConstant = 20000;
#endif
	if( !::SetCommTimeouts( m_hPort, &TimeOut ) ){
		::CloseHandle( m_hPort );
		m_hPort = INVALID_HANDLE_VALUE;
		return FALSE;
	}
	::GetCommState( m_hPort, &m_dcb );
	m_dcb.BaudRate = pPara->m_Baud;
	m_dcb.ByteSize = BYTE(pPara->m_Length ? 8 : 7);
	m_dcb.StopBits = BYTE(pPara->m_Stop ? TWOSTOPBITS : ONESTOPBIT);
	m_dcb.fBinary = TRUE;
	m_dcb.Parity = BYTE(pPara->m_Parity);
	m_dcb.XonChar = 0x11;	// XON
	m_dcb.XoffChar = 0x13;	// XOFF
	m_dcb.fParity = pPara->m_Parity != 0;
	m_dcb.fOutxDsrFlow = FALSE;
	m_dcb.EvtChar = 0x0d;
	m_dcb.fDtrControl = (pPara->m_PTT & 1 ) ? DTR_CONTROL_DISABLE : DTR_CONTROL_ENABLE;
	m_dcb.fRtsControl = (pPara->m_PTT & 2 ) ? RTS_CONTROL_DISABLE : RTS_CONTROL_ENABLE;
	m_dcb.fOutxCtsFlow = ((pPara->m_Flow & 2) && !pPara->m_PTT) ? TRUE : FALSE;
	m_dcb.fInX = m_dcb.fOutX = (pPara->m_Flow & 1) ? TRUE : FALSE;
	m_dcb.fTXContinueOnXoff = FALSE;
	m_dcb.XonLim = USHORT(256/3);
	m_dcb.XoffLim = USHORT(256*2/3);
	m_dcb.DCBlength = sizeof( DCB );
	if( !::SetCommState( m_hPort, &m_dcb ) ){
		::CloseHandle( m_hPort );
		m_hPort = INVALID_HANDLE_VALUE;
		return FALSE;
	}
	::SetCommMask( m_hPort, EV_RXFLAG );

	m_pThread = AfxBeginThread(CommThreadProc, LPVOID(this));
	ASSERT(m_pThread);
	if( m_pThread ) m_pThread->m_bAutoDelete = FALSE;
	return TRUE;
}

//---------------------------------------------------------------------------
void __fastcall CComm::Close(void)
{
	if( m_pThread ){
		m_Terminated = TRUE;
		::WaitForSingleObject(m_pThread->m_hThread, 3000);
		delete m_pThread;
		m_pThread = NULL;
	}
	if( m_hPort != INVALID_HANDLE_VALUE ){
		::CloseHandle(m_hPort);
		m_hPort = INVALID_HANDLE_VALUE;
	}
}
//---------------------------------------------------------------------------
void __fastcall CComm::SetPTT(int sw)
{
	if( m_pPara == NULL ) return;
	if( m_hPort == INVALID_HANDLE_VALUE ) return;

	int ct = m_pPara->m_PTT;
	if( ct & 1 ) ::EscapeCommFunction(m_hPort, sw ? SETDTR : CLRDTR);
	if( ct & 2 ) ::EscapeCommFunction(m_hPort, sw ? SETRTS : CLRRTS);
}
//---------------------------------------------------------------------------
void __fastcall CComm::PutChar(BYTE c)
{
	if( m_hPort == INVALID_HANDLE_VALUE ) return;
	if( m_FifoTX.IsBufferFull() ) return;

	m_FifoTX.PutData(c);
}

//---------------------------------------------------------------------------
BYTE __fastcall CComm::GetChar(void)
{
	if( m_hPort == INVALID_HANDLE_VALUE ) return 0;
	if( m_FifoRX.GetCount() ){
		return m_FifoRX.GetData();
	}
	else {
		return 0;
	}
}

//---------------------------------------------------------------------------
int __fastcall CComm::GetRxLen(void)
{
	return m_FifoRX.GetCount();
}

//---------------------------------------------------------------------------
UINT CommThreadProc(LPVOID pParam)
{
	CComm *pComm = (CComm *)pParam;
	while(!pComm->m_Terminated){
		COMSTAT ComStat;
		DWORD	dwErrorFlags;
		::ClearCommError( pComm->m_hPort, &dwErrorFlags, &ComStat );
#if USE_POLLING
		if( ComStat.cbInQue ){  		// RX data exist.
			BYTE bf[256];
			DWORD i = ComStat.cbInQue;
			if( i > sizeof(bf) ) i = sizeof(bf);
			DWORD size = 0;
			::ReadFile(pComm->m_hPort, bf, i, &size, NULL );
			for( i = 0; i < size; i++ ) pComm->m_FifoRX.PutData(bf[i]);
			::PostMessage(pComm->m_hWnd, WM_COMMEVENT, 0, 0);
		}
		if( pComm->m_FifoTX.GetCount() ){		// TX data exist
			if( !ComStat.cbOutQue ){    	// Is the TX-buffer empty?
				DWORD size = 0;
				::WriteFile(pComm->m_hPort, pComm->m_FifoTX.GetP(), 1, &size, NULL );
				if( size ){
					pComm->m_FifoTX.GetData();
				}
			}
			else {
				::Sleep(1);
			}
		}
		else {
			::Sleep(10);
		}
#else
		if( !pComm->m_OverlapTX.hEvent ){
			if( pComm->m_FifoTX.GetCount() ){
				pComm->m_TxD = pComm->m_FifoTX.GetData();
				pComm->m_OverlapTX.hEvent = (HANDLE)pComm;
				::WriteFileEx(pComm->m_hPort, &pComm->m_TxD, 1, &pComm->m_OverlapTX, FileIOCompletionTX);
			}
		}
		if( !pComm->m_OverlapRX.hEvent ){
			pComm->m_OverlapRX.hEvent = (HANDLE)pComm;
			::ReadFileEx(pComm->m_hPort, &pComm->m_RxD, 1, &pComm->m_OverlapRX, FileIOCompletionRX);
		}
		::SleepEx(20, TRUE);
#endif
	}
	return 0;
}
#if !USE_POLLING
//---------------------------------------------------------------------------
VOID WINAPI FileIOCompletionTX(DWORD fdwError, DWORD cbTransferred, LPOVERLAPPED lpo)
{
	lpo->hEvent = NULL;
}
//---------------------------------------------------------------------------
VOID WINAPI FileIOCompletionRX(DWORD fdwError, DWORD cbTransferred, LPOVERLAPPED lpo)
{
	if( cbTransferred ){
		CComm *pComm = (CComm *)lpo->hEvent;
		pComm->m_FifoRX.PutData(pComm->m_RxD);
		::PostMessage(pComm->m_hWnd, WM_COMMEVENT, 0, 0);
	}
	lpo->hEvent = NULL;
}
#endif
//---------------------------------------------------------------------------
